/*
 *
 * ***** BEGIN GPL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * ***** END GPL LICENSE BLOCK *****
 *
 * Contributor(s): Jiri Hnidek <jiri.hnidek@tul.cz>.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <signal.h>
#include <getopt.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <wiringPi.h>
#include <wiringPiI2C.h>

static int running = 0;
static int sleep_delay = 1;
static int daemon_mode = 0;
static char *conf_file_name = NULL;
static char *pid_file_name = NULL;
static int pid_fd = -1;
static FILE *usb_pwr_file = NULL;
static char *app_name = NULL;
static FILE *log_stream;
static int i2c_dev;
static int daemon_state = -1;
char *buffer;
char *oad_buffer;
char *oad_buffer_2;


/**
 * \brief Read deamon mode
 */
void set_daemon_mode(void)
{
	FILE *daemon_mode_file = NULL;
	int ret = -1;

	daemon_mode_file = fopen("/home/pi/daemon_mode", "r");

	if (daemon_mode_file == NULL) {
		daemon_mode = 0;
		return;
	}

	ret = fscanf(daemon_mode_file, "%d", &daemon_mode);
	if (ret > 0) {
		if (daemon_mode != 1) {
			daemon_mode = 0;
		} else {
			daemon_mode = 1;
		}
	}

	fclose(daemon_mode_file);
}

/**
 * \brief Read configuration from config file
 */
int read_conf_file(int reload)
{
	FILE *conf_file = NULL;
	int ret = -1;

	if (conf_file_name == NULL) return 0;

	conf_file = fopen(conf_file_name, "r");

	if (conf_file == NULL) {
		syslog(LOG_ERR, "Can not open config file: %s, error: %s",
				conf_file_name, strerror(errno));
		return -1;
	}

	ret = fscanf(conf_file, "%d", &sleep_delay);

	if (ret > 0) {
		if (reload == 1) {
			syslog(LOG_INFO, "Reloaded configuration file %s of %s",
				conf_file_name,
				app_name);
		} else {
			syslog(LOG_INFO, "Configuration of %s read from file %s",
				app_name,
				conf_file_name);
		}
	}

	fclose(conf_file);

	return ret;
}

/**
 * \brief This function tries to test config file
 */
int test_conf_file(char *_conf_file_name)
{
	FILE *conf_file = NULL;
	int ret = -1;

	conf_file = fopen(_conf_file_name, "r");

	if (conf_file == NULL) {
		fprintf(stderr, "Can't read config file %s\n",
			_conf_file_name);
		return EXIT_FAILURE;
	}

	ret = fscanf(conf_file, "%d", &sleep_delay);

	if (ret <= 0) {
		fprintf(stderr, "Wrong config file %s\n",
			_conf_file_name);
	}

	fclose(conf_file);

	if (ret > 0)
		return EXIT_SUCCESS;
	else
		return EXIT_FAILURE;
}

/**
 * \brief Callback function for handling signals.
 * \param	sig	identifier of signal
 */
void handle_signal(int sig)
{
	if (sig == SIGINT) {
		fprintf(log_stream, "Debug: stopping daemon ...\n");
		/* Unlock and close lockfile */
		if (pid_fd != -1) {
			lockf(pid_fd, F_ULOCK, 0);
			close(pid_fd);
		}
		/* Try to delete lockfile */
		if (pid_file_name != NULL) {
			unlink(pid_file_name);
		}
		running = 0;
		/* Reset signal handling to default behavior */
		signal(SIGINT, SIG_DFL);
	} else if (sig == SIGHUP) {
		fprintf(log_stream, "Debug: reloading daemon config file ...\n");
		read_conf_file(1);
	} else if (sig == SIGCHLD) {
		fprintf(log_stream, "Debug: received SIGCHLD signal\n");
	}
}

/**
 * \brief This function will daemonize this app
 */
static void daemonize()
{
	pid_t pid = 0;
	int fd;

	/* Fork off the parent process */
	pid = fork();

	/* An error occurred */
	if (pid < 0) {
		exit(EXIT_FAILURE);
	}

	/* Success: Let the parent terminate */
	if (pid > 0) {
		exit(EXIT_SUCCESS);
	}

	/* On success: The child process becomes session leader */
	if (setsid() < 0) {
		exit(EXIT_FAILURE);
	}

	/* Ignore signal sent from child to parent process */
	signal(SIGCHLD, SIG_IGN);

	/* Fork off for the second time*/
	pid = fork();

	/* An error occurred */
	if (pid < 0) {
		exit(EXIT_FAILURE);
	}

	/* Success: Let the parent terminate */
	if (pid > 0) {
		exit(EXIT_SUCCESS);
	}

	/* Set new file permissions */
	umask(0);

	/* Change the working directory to the root directory */
	/* or another appropriated directory */
	chdir("/");

	/* Close all open file descriptors */
	for (fd = sysconf(_SC_OPEN_MAX); fd > 0; fd--) {
		close(fd);
	}

	/* Reopen stdin (fd = 0), stdout (fd = 1), stderr (fd = 2) */
	stdin = fopen("/dev/null", "r");
	stdout = fopen("/dev/null", "w+");
	stderr = fopen("/dev/null", "w+");

	/* Try to write PID of daemon to lockfile */
	if (pid_file_name != NULL)
	{
		char str[256];
		pid_fd = open(pid_file_name, O_RDWR|O_CREAT, 0640);
		if (pid_fd < 0) {
			/* Can't open lockfile */
			exit(EXIT_FAILURE);
		}
		if (lockf(pid_fd, F_TLOCK, 0) < 0) {
			/* Can't lock file */
			exit(EXIT_FAILURE);
		}
		/* Get current PID */
		sprintf(str, "%d\n", getpid());
		/* Write PID to lockfile */
		write(pid_fd, str, strlen(str));
	}
}

/**
 * \brief Print help for this application
 */
void print_help(void)
{
	printf("\n Usage: %s [OPTIONS]\n\n", app_name);
	printf("  Options:\n");
	printf("   -h --help                 Print this help\n");
	printf("   -c --conf_file filename   Read configuration from the file\n");
	printf("   -t --test_conf filename   Test configuration file\n");
	printf("   -l --log_file  filename   Write logs to the file\n");
	printf("   -d --daemon               Daemonize this application\n");
	printf("   -p --pid_file  filename   PID file used by daemonized app\n");
	printf("   -b --battery_voltage   	 Read battery voltage from battery controller\n");
	printf("   -u --usb_voltage   	     Read USB input voltage from battery controller\n");
	printf("   -u --usb_voltage   	     Read USB input voltage from battery controller\n");
	printf("\n");
}

int read_oad_header(char *name, char *buffer)
{
	FILE *file;
	size_t byteCnt;

	//Open file
	file = fopen(name, "rb");
	if (!file)
	{
		printf("Unable to open file %s\n", name);
		return -1;
	}

	//Get file length
	fseek(file, 0, SEEK_END);
	byteCnt = ftell(file);

	//Rewind
	fseek(file, 0, SEEK_SET);

	//Read file contents into buffer
	fread(buffer, 16, 1, file);

	fclose(file);
	
	return byteCnt;
}


int read_oadc_block(char *name, int buffer_index, char *oad_buffer, int block_length) {
	FILE *file;
	//size_t byteCnt;
	int i;

	//Open file
	file = fopen(name, "rb");
	if (!file)
	{
		printf("Unable to open file %s\n", name);
		return -1;
	}

	//Get file length
	//fseek(file, 0, SEEK_END);
	//byteCnt = ftell(file);
	
	//Set file pointer to buffer start
	fseek(file, buffer_index*64, SEEK_SET);

	//Read file contents into buffer
	fread(oad_buffer, block_length, 1, file);
	
	if (block_length < 64) {
		// Fill rest with FFs 
		for (i=block_length; i<64; i++) {
			oad_buffer[i] = 255;
		}
	}

	fclose(file);
	
	return 1;
}

void insert_block_header(char *oad_buffer,char *oad_buffer_2, int oadc) {
	
	int i;
	
	oad_buffer_2[1] = oadc % 256;
	oad_buffer_2[0] = oadc / 256;
	
	for (i=0;i<64;i++) {
		oad_buffer_2[2+i] = oad_buffer[i];
	}
}


void dump_buffer(void *buffer, int buffer_size)
{
  int i;

  for(i = 0;i < buffer_size;++i) {
     printf("%02X ",(int)((char *)buffer)[i]);
  }
}


/* Main function */
int main(int argc, char *argv[])
{
	static struct option long_options[] = {
		{"conf_file", required_argument, 0, 'c'},
		{"test_conf", required_argument, 0, 't'},
		{"log_file", required_argument, 0, 'l'},
		{"ota_update", required_argument,0,'o'},
		{"battery_voltage", no_argument, 0, 'b'},
		{"usb_voltage", no_argument, 0, 'u'},
		{"ir_state", required_argument, 0, 'i'},
		{"help", no_argument, 0, 'h'},
		{"daemon", no_argument, 0, 'd'},
		{"pid_file", required_argument, 0, 'p'},
		{NULL, 0, 0, 0}
	};
	int value, option_index = 0;
	char *log_file_name = NULL;
	char *irda_control_state = NULL;
	int start_daemonized = 0;
	int i2c_val = 0;
	int wpPin9val = 0;
	int wpPin15val = 0;
	int wpPin12val = 0; // power switch input
	int wpPin8val = 0;
	int bc=0;
	int oadc = 0;
	int oadbc = 0;
	int oad_delay_cntr = 0;
	int block_count = 0;
	int last_block_bc = 0;
	int current_block_length = 64;
	
	app_name = argv[0];

	/* Try to process all command line arguments */
	while ((value = getopt_long(argc, argv, "c:l:t:i:p:o:budh", long_options, &option_index)) != -1) {
		switch (value) {
			case 'c':
				conf_file_name = strdup(optarg);
				break;
			case 'l':
				log_file_name = strdup(optarg);
				break;
			case 'p':
				pid_file_name = strdup(optarg);
				break;
			case 't':
				return test_conf_file(optarg);
			case 'i':
				irda_control_state = strdup(optarg);
				wiringPiSetup();
				pinMode(0,OUTPUT); // WP PIN 0 IS OUTPUT
				if ((int)irda_control_state[0] == 49) {
					printf("Writing 1 to IR output\n");
					digitalWrite(0,1);
				} else {
					printf("Writing 0 to IR output\n");
					digitalWrite(0,0);
				}
				return EXIT_SUCCESS;
			case 'o':
				set_daemon_mode();
				if (daemon_mode == 1) {
					buffer=(char *)malloc(16);
					oad_buffer=(char *)malloc(64);
					oad_buffer_2=(char *)malloc(66);
					if (!buffer || !oad_buffer  || !oad_buffer_2)
					{
						printf("Memory failure\n");
						if (buffer) free(buffer);
						if (oad_buffer) free(oad_buffer);
						if (oad_buffer_2) free(oad_buffer_2);
						return EXIT_FAILURE;
					} else {
						bc = read_oad_header(strdup(optarg),buffer);
						block_count = bc / 64;
						last_block_bc = bc % 64;
						if (last_block_bc > 0) {
							block_count = block_count+1;
						}
						printf("OAD file byte count: %d\n",bc);
						printf("OAD file total block count: %d\n",block_count);
						printf("OAD file overflow byte count: %d\n",last_block_bc);
						if (bc >16) {
							dump_buffer(buffer,16);
							printf("\n");
							i2c_dev = wiringPiI2CSetup(0x48) ;
							if (i2c_dev < 0) {
								printf("Can not set up I2C for address 0x48\n");
								free(buffer);
								free(oad_buffer);
								free(oad_buffer_2);
								exit(EXIT_FAILURE);
							} else {
								for (oadc=0; oadc<16; oadc++) {
									//i2c_val =  wiringPiI2CWriteReg16(i2c_dev,6,(int)buffer[oadc]);
									i2c_val =  wiringPiI2CWriteReg8(i2c_dev,6,(int)buffer[oadc]);
									if (i2c_val < 0) {
										printf("Could not write to register 6\n");
										free(buffer);
										free(oad_buffer);
										free(oad_buffer_2);
										exit(EXIT_FAILURE);
									}
								}
								i2c_val =  wiringPiI2CWriteReg8(i2c_dev,5,1);
								if (i2c_val < 0) {
									printf("Could not write to register 5\n");
									free(buffer);
									free(oad_buffer);
									free(oad_buffer_2);
									exit(EXIT_FAILURE);
								} else {
									for (oad_delay_cntr=0;oad_delay_cntr<10;oad_delay_cntr++) {
										usleep(100000);
										//i2c_val =  wiringPiI2CReadReg16(i2c_dev,7);
										i2c_val =  wiringPiI2CReadReg8(i2c_dev,7);
										printf("%d\n",i2c_val);
										if (i2c_val == 1) {
											break;
										}
										if (i2c_val == 2) {
											printf("Image header rejected\n");
											free(buffer);
											free(oad_buffer);
											free(oad_buffer_2);
											exit(EXIT_FAILURE);
										}
									}
									if (i2c_val != 1) {
										printf("Waiting for ack to image header check\n");
										free(buffer);
										free(oad_buffer);
										free(oad_buffer_2);
										exit(EXIT_FAILURE);
									}
								}
							}
							if (bc < 16) {
								printf("File too short: %d: bytes\n",bc);
							} else {
								for (oadc=0;oadc<block_count;oadc++) {
									if (oadc == (block_count -1 )) {
											if (last_block_bc > 0) {
												current_block_length = last_block_bc;
											} else {
												current_block_length = 64;
											}
									} else {
										current_block_length = 64;
									}
									if (read_oadc_block(strdup(optarg),oadc,oad_buffer,current_block_length)) {
										insert_block_header(oad_buffer,oad_buffer_2,oadc);
										dump_buffer(oad_buffer_2,66);
										printf("\n");
										
										for (oadbc=0;oadbc<66;oadbc++) {
	
											i2c_val =  wiringPiI2CWriteReg8(i2c_dev,6,(int)oad_buffer_2[oadbc]);
											if (i2c_val < 0) {
												printf("Could not write oad buffer to register 6\n");
												free(buffer);
												free(oad_buffer);
												free(oad_buffer_2);
												exit(EXIT_FAILURE);
											}
										}
										i2c_val =  wiringPiI2CWriteReg8(i2c_dev,5,2);
										if (i2c_val < 0) {
											printf("Could not write to register 5\n");
											free(buffer);
											free(oad_buffer);
											free(oad_buffer_2);
											exit(EXIT_FAILURE);
										} else {
											for (oad_delay_cntr=0;oad_delay_cntr<1000;oad_delay_cntr++) {
												//usleep(1000000);
												usleep(30000);
												i2c_val =  wiringPiI2CReadReg8(i2c_dev,7);
												//printf("%d\n",i2c_val);
												if ((i2c_val == 1) || (i2c_val == 2)) {
													break;
												}
											}
											if ((oad_delay_cntr==10) || (i2c_val == 2)) {
												printf("Timeout or NACK to oad buffer write\n");
												free(buffer);
												free(oad_buffer);
												free(oad_buffer_2);
												exit(EXIT_FAILURE);
											}
										}
										
									} else {
										printf("Failed to read buffer %d from file: \n",oadc);
										free(buffer);
										free(oad_buffer);
										free(oad_buffer_2);
										return EXIT_FAILURE;
									}
								}
							}
							printf("File length %d\n",bc);
						} else {
							printf("Too short file %d: \n",bc);
							free(buffer);
							free(oad_buffer);
							free(oad_buffer_2);
							return EXIT_FAILURE;
						}
						free(buffer);
						free(oad_buffer);
						free(oad_buffer_2);
					}
					return EXIT_SUCCESS;
				} else {
					printf("Command not supported in mode 0\n");
					return EXIT_SUCCESS;
				}
			case 'd':
				start_daemonized = 1;
				break;
			case 'b':
				set_daemon_mode();
				if (daemon_mode == 1) {
					i2c_dev = wiringPiI2CSetup(0x48) ;
					i2c_val =  wiringPiI2CReadReg16(i2c_dev,3);
					if (i2c_val < 0) {
						printf("Failed to read from I2C address %s, register%d\n", "0x48", 3);
					} else {
						printf("%d\n", i2c_val);
					}
					return EXIT_SUCCESS;
				} else {
					printf("Command not supported in mode 0\n");
					return EXIT_SUCCESS;
				}
			case 'u':
				set_daemon_mode();
				if (daemon_mode == 1) {
					i2c_dev = wiringPiI2CSetup(0x48) ;
					i2c_val =  wiringPiI2CReadReg16(i2c_dev,4);
					if (i2c_val < 0) {
						printf("Failed to read from I2C address %s, register%d\n", "0x48", 4);
					} else {
						printf("%d\n", i2c_val);
					}
					return EXIT_SUCCESS;
				} else {
					printf("Command not supported in mode 0\n");
					return EXIT_SUCCESS;					
				}
			case 'h':
				print_help();
				return EXIT_SUCCESS;
			case '?':
				print_help();
				return EXIT_FAILURE;
			default:
				break;
		}
	}

	/* When daemonizing is requested at command line. */
	if (start_daemonized == 1) {
		/* It is also possible to use glibc function deamon()
		 * at this point, but it is useful to customize your daemon. */
		daemonize();
	}

	/* Open system log and write message to it */
	openlog(argv[0], LOG_PID|LOG_CONS, LOG_DAEMON);
	syslog(LOG_INFO, "Started %s", app_name);

	/* Daemon will handle two signals */
	signal(SIGINT, handle_signal);
	signal(SIGHUP, handle_signal);

	/* Try to open log file to this daemon */
	if (log_file_name != NULL) {
		log_stream = fopen(log_file_name, "a+");
		if (log_stream == NULL) {
			syslog(LOG_ERR, "Can not open log file: %s, error: %s",
				log_file_name, strerror(errno));
			log_stream = stdout;
		}
	} else {
		log_stream = stdout;
	}

	/* Read configuration from config file */
	read_conf_file(0);

	/* This global variable can be changed in function handling signal */
	running = 1;
		
	// Open usb power status file and write initial value to it
	usb_pwr_file = fopen("/home/pi/usb_pwr", "w+");
	if (usb_pwr_file == NULL) {
		/* Can't open usb power monitoring file*/
		fprintf(log_stream, "Failed to open usb power monitoring file\n");
	} else {
		char str[100];
		sprintf(str, "%d",1);
		fwrite(str , 1 , 1 , usb_pwr_file );		
	}
	
	set_daemon_mode();
	if (daemon_mode == 0) {
		// INITIALIZE HERE GPIO MODE FOR WIRING PI
		wiringPiSetup();
		pinMode(12,INPUT);
		pullUpDnControl (12,PUD_OFF);
		pinMode(9,INPUT); // WP PIN 9 IS INPUT
		pullUpDnControl (9,PUD_DOWN);
		pinMode(8,OUTPUT); // WP PIN 8 IS OUTPUT
		pullUpDnControl (8,PUD_DOWN);
		digitalWrite(8,0);
		pinMode(15,INPUT); // WP PIN 15 IS INPUT
		pinMode(27,OUTPUT); // WP PIN 27 IS OUTPUT
		digitalWrite(27,1);

	} else {
		i2c_dev = wiringPiI2CSetup(0x48) ;
		if (i2c_dev < 0) {
			syslog(LOG_ERR, "Can not set up I2C for address 0x48: %s, error: %s",
				(log_stream == stdout) ? "stdout" : log_file_name, strerror(errno));
			exit(EXIT_FAILURE);
		}
		
		i2c_val =  wiringPiI2CWriteReg16(i2c_dev,1,0);
		if (i2c_val < 0) {
			fprintf(log_stream, "Could not initialize I2C address 0x48 register %d\n", 1);
		}
	}
	
	/* Never ending loop of server */	
	while (running == 1) {
		if (daemon_mode == 1) {
			// Deamon uses i2c interface to determine power state
			sleep(sleep_delay);
			switch (daemon_state) {
			case -1:
				fprintf(log_stream, "Started deamon loop\n");
				daemon_state = 0;
				break;
			case 0:
				i2c_val =  wiringPiI2CReadReg8(i2c_dev,2);
				//fprintf(log_stream, "Reg 2 value: %d\n", i2c_val);
				if (i2c_val < 0) {
					fprintf(log_stream, "Failed to read from I2C address %s, register%d\n", "0x48", 2);
				} else {
					if (usb_pwr_file != NULL) {
						char rstr[100];
						rewind(usb_pwr_file);
						fread(rstr, 1, 1, usb_pwr_file);
						char str[100];
						sprintf(str, "%d",i2c_val);
						if (str[0] != rstr[0]) {
							fprintf(log_stream, "USB power state has changed from %s to %s\n", rstr,str);
							rewind(usb_pwr_file);
							fwrite(str , 1 , 1 , usb_pwr_file );
						}
					}
				}
				i2c_val =  wiringPiI2CReadReg8(i2c_dev,1);		
				if (i2c_val < 0) {
					fprintf(log_stream, "Failed to read from I2C address %s, register%d\n", "0x48", 1);
				} else {
					//fprintf(log_stream, "Reg 1 value: %d\n", i2c_val);
					if (i2c_val == 1) {
						i2c_val =  wiringPiI2CWriteReg16(i2c_dev,1,0);
						daemon_state = 1; // Shutting down
						fprintf(log_stream, "POWER OFF COMMAND: SHUTTING DOWN RASPBERRY PI");
						system("sudo shutdown -h now");
					}
				}
				break;
			case 1:
				// DO nothing - system is shutting down.
				break;
			}
		} else {
			// Deamon uses GPIO pin protocol to determine power state
			sleep(sleep_delay);
			if (wpPin8val == 0) {
				wpPin8val = 1;
			} else {
				wpPin8val = 0;
			}
			switch (daemon_state) {
			case -1:
				fprintf(log_stream, "Started deamon loop\n");
				daemon_state = 0;
				break;
			case 0:
				wpPin15val = digitalRead(15);
				//fprintf(log_stream, "Reg 2 value: %d\n", i2c_val);
				if (usb_pwr_file != NULL) {
					char rstr[1];
					rewind(usb_pwr_file);
					fread(rstr, 1, 1, usb_pwr_file);
					char str[1];
					sprintf(str, "%d",wpPin15val);
					if (str[0] != rstr[0]) {
						fprintf(log_stream, "USB power state has changed from %s to %s\n", rstr,str);
						rewind(usb_pwr_file);
						fwrite(str , 1 , 1 , usb_pwr_file );
					}
				}
				
				wpPin12val =  digitalRead(12);	
				fprintf(log_stream, "Power switch value: %d\n", wpPin12val);
				
				wpPin9val =  digitalRead(9);	
				fprintf(log_stream, "WP PIN 9 value: %d\n", wpPin9val);
				if (wpPin9val == 1) {
					digitalWrite(8,1); // Ack the Power off request
					daemon_state = 1; // Shutting down
					fprintf(log_stream, "POWER OFF COMMAND: SHUTTING DOWN RASPBERRY PI");
					system("sudo shutdown -h now");
				} else {
					digitalWrite(8,wpPin8val); // Alive signal
				}
				break;
			case 1:
				// DO nothing - system is shutting down.
				wpPin9val =  digitalRead(9);
				fprintf(log_stream, "WP PIN 9 value: %d\n", wpPin9val);
				break;
			}			
		}
	}

	/* Close log file, when it is used. */
	if (log_stream != stdout) {
		fclose(log_stream);
	}

	/* Write system log and close it. */
	syslog(LOG_INFO, "Stopped %s", app_name);
	closelog();

	/* Free allocated memory */
	if (conf_file_name != NULL) free(conf_file_name);
	if (log_file_name != NULL) free(log_file_name);
	if (pid_file_name != NULL) free(pid_file_name);

	return EXIT_SUCCESS;
}
